/*
 * FAS216 generic driver
 *
 * Copyright (C) 1997 Russell King
 */
#ifndef FAS216_H
#define FAS216_H

#ifndef NO_IRQ
#define NO_IRQ 255
#endif

#include "queue.h"
#include "msgqueue.h"

/* FAS register definitions */

/* transfer count low */
#define REG0_LOCOUNT		(0)

/* transfer count medium */
#define REG0_MDCOUNT		(1)

/* fifo data */
#define REG0_FIFODATA		(2)

/* command */
#define REG0_CMD		(3)
#define CMD_NOP			0x00
#define CMD_FLUSHFIFO		0x01
#define CMD_RESETCHIP		0x02
#define CMD_RESETSCSI		0x03

#define CMD_TRANSFERINFO	0x10
#define CMD_INITCMDCOMPLETE	0x11
#define CMD_MSGACCEPTED		0x12
#define CMD_SETATN		0x1a
#define CMD_RSETATN		0x1b

#define CMD_SELECTWOATN		0x41
#define CMD_SELECTATN		0x42
#define CMD_SELECTATNSTOP	0x43
#define CMD_ENABLESEL		0x44
#define CMD_DISABLESEL		0x45
#define CMD_SELECTATN3		0x46
#define CMD_RESEL3		0x47

#define CMD_WITHDMA		0x80

/* status register (read) */
#define REG0_STATUS		(4)
#define STATUS_IO		(1 << 0)			/* IO phase		*/
#define STATUS_CD		(1 << 1)			/* CD phase		*/
#define STATUS_MSG		(1 << 2)			/* MSG phase		*/
#define STATUS_TRANSFERDONE	(1 << 3)			/* Transfer completed	*/
#define STATUS_TRANSFERCNTZ	(1 << 4)			/* Transfer counter is zero */
#define STATUS_PARITYERROR	(1 << 5)			/* Parity error		*/
#define STATUS_REALBAD		(1 << 6)			/* Something bad	*/
#define STATUS_INT		(1 << 7)			/* Interrupt		*/

#define STATUS_BUSMASK		(STATUS_MSG|STATUS_CD|STATUS_IO)
#define STATUS_DATAOUT		(0)				/* Data out		*/
#define STATUS_DATAIN		(STATUS_IO)			/* Data in		*/
#define STATUS_COMMAND		(STATUS_CD)			/* Command out		*/
#define STATUS_STATUS		(STATUS_CD|STATUS_IO)		/* Status In		*/
#define STATUS_MESGOUT		(STATUS_MSG|STATUS_CD)		/* Message out		*/
#define STATUS_MESGIN		(STATUS_MSG|STATUS_CD|STATUS_IO)/* Message In		*/

/* bus ID for select / reselect */
#define REG0_BUSID		(4)
#define BUSID(target)		((target) & 7)

/* Interrupt status register (read) */
#define REG0_ISR		(5)
#define ISR_SELWOATN		(1 << 0)			/* Select w/o ATN	*/
#define ISR_SELATN		(1 << 1)			/* Select w/ATN		*/
#define ISR_RESELECTED		(1 << 2)			/* Reselected		*/
#define ISR_FUNCDONE		(1 << 3)			/* Function done	*/
#define ISR_BUSSERVICE		(1 << 4)			/* Bus service		*/
#define ISR_DISCONNECT		(1 << 5)			/* Disconnect		*/
#define ISR_ILLEGALCMD		(1 << 6)			/* Illegal command	*/
#define ISR_BUSRESET		(1 << 7)			/* SCSI Bus reset	*/

/* Timeout register (write) */
#define REG0_TIMEOUT		(5)

/* Sequence step register (read) */
#define REG0_SSR		(6)
#define SSR_BITS		0x07
#define SSR_SELARB		0x00				/* Select & Arb ok	*/
#define SSR_MSGBYTESENT		0x01				/* One byte message sent*/
#define SSR_NOTCOMMAND		0x02				/* Not in command state	*/
#define SSR_EARLYPHASE		0x03				/* Early phase change	*/
#define SSR_COMPLETE		0x04				/* Command ok		*/

/* Transfer period step (write) */
#define REG0_PERIOD		(6)

/* Synchronous Offset (write) */
#define REG0_SYNCOFFSET		(7)

/* Fifo state register (read) */
#define REG0_FIFOSTATE		(7)
#define FF_BYTES		0x1f				/* Num bytes in FIFO	*/
#define FF_SSTEP		0xe0				/* Step			*/

/* config register 1 */
#define REG0_CFG1		(8)
#define CFG1_OWNID		(7 << 0)			/* Own ID		*/
#define CFG1_ENABLETEST		(1 << 3)			/* Enable tests		*/
#define CFG1_ENABLEPARITY	(1 << 4)			/* Enable parity	*/
#define CFG1_PARTYTEST		(1 << 5)			/* Parity test disable	*/
#define CFG1_DISABLERESETREP	(1 << 6)			/* Disable SCSI reset	*/
#define CFG1_SLOWCABLE		(1 << 7)			/* Enable slow cable	*/

/* Clock conversion factor (read) */
#define REG0_CLOCKCONV		(9)	/* 00 */
#define CCF_F37MHZ		0x00				/* 35 - 40 MHz		*/
#define CCF_F10MHZ		0x02				/* 10 MHz		*/
#define CCF_F12MHZ		0x03				/* 10 - 15 MHz		*/
#define CCF_F17MHZ		0x04				/* 15 - 20 MHz		*/
#define CCF_F22MHZ		0x05				/* 20 - 25 MHz		*/
#define CCF_F27MHZ		0x06				/* 25 - 30 MHz		*/
#define CCF_F32MHZ		0x07				/* 30 - 35 MHz		*/

/* Chip test register (write) */
#define REG0_TEST		(10)
#define TEST_TARG		0x01				/* Target test mode	*/
#define TEST_INI		0x02				/* Initiator test mode	*/
#define TEST_TRISTATE		0x04				/* Tristate test mode	*/

/* Configuration register 2 (read/write) */
#define REG0_CFG2		(11)	/* 48 */
#define CFG2_DMAPARITY		(1 << 0)			/* Parity DMA error	*/
#define CFG2_REGPARITY		(1 << 1)			/* Parity Register Err	*/
#define CFG2_BADPARITY		(1 << 2)			/* Bad parity targ.err	*/
#define CFG2_SCSI2ENABLE	(1 << 3)			/* Enable SCSI2		*/
#define CFG2_HIDRQ		(1 << 4)			/* Tristate DREQ	*/
#define CFG2_BYTECTRL		(1 << 5)			/* Byte Control		*/
#define CFG2_FENABLE		(1 << 6)			/* Enable features	*/

/* Configuration register 3 (read/write) */
#define REG0_CFG3		(12)	/* 38 */
#define CFG3_SAVERESIDUALBYTE	(1 << 2)
#define CFG3_GROUP2		(1 << 5)			/* Group2 SCSI support	*/
#define CFG3_3BYTEMSGS		(1 << 6)			/* 3 byte msgs ok	*/
#define CFG3_IDENABLE		(1 << 7)			/* ID message checking	*/

/* High transfer count (read/write) */
#define REG0_HICOUNT		(14)

/* ID reigster (read only) */
#define REG1_ID			(14)

/* Data alignment */
#define REG0_DATAALIGN		(15)

typedef enum {
    PHASE_IDLE,					/* we're not planning on doing anything	*/
    PHASE_SELECTION,				/* selecting a device			*/
    PHASE_RECONNECTED,				/* reconnected				*/
    PHASE_DATAOUT,				/* data out to device			*/
    PHASE_DATAIN,				/* data in from device			*/
    PHASE_MSGIN,				/* message in from device		*/
    PHASE_MSGOUT,				/* message out to device		*/
    PHASE_AFTERMSGOUT,				/* after message out phase		*/
    PHASE_STATUS,				/* status from device			*/
    PHASE_DISCONNECT,				/* disconnecting from bus		*/
    PHASE_DONE					/* Command complete			*/
} phase_t;

typedef enum {
    DMA_OUT,					/* DMA from memory to chip		*/
    DMA_IN					/* DMA from chip to memory		*/
} dmadir_t;

typedef struct {
    struct Scsi_Host	*host;			/* host					*/
    Scsi_Cmnd		*SCpnt;			/* currently processing command		*/
    Scsi_Cmnd		*origSCpnt;		/* original connecting command		*/

    /* driver information */
    struct {
	unsigned int	io_port;		/* base address of FAS216		*/
	unsigned char	irq;			/* interrupt				*/
	unsigned char	cfg[4];			/* configuration registers		*/
	const char	*type;			/* chip type				*/
	phase_t		phase;			/* current phase			*/

	struct {
		unsigned char	target;		/* reconnected target			*/
		unsigned char	lun;		/* reconnected lun			*/
		unsigned char	tag;		/* reconnected tag			*/
	} reconnected;

	Scsi_Pointer	SCp;			/* current commands data pointer	*/

	MsgQueue_t	msgs;			/* message queue for connected device	*/

	unsigned short	last_message;		/* last message to be sent		*/

	unsigned char	disconnectable:1;	/* this command can be disconnected	*/
	unsigned char	aborting:1;		/* aborting command			*/
    } scsi;

    /* statistics information */
    struct {
	unsigned int	queues;
	unsigned int	removes;
	unsigned int	fins;
	unsigned int	reads;
	unsigned int	writes;
	unsigned int	miscs;
	unsigned int	disconnects;
	unsigned int	aborts;
	unsigned int	resets;
    } stats;

    /* configuration information */
    struct {
	unsigned char	clockrate;		/* clock rate of FAS device (MHz)	*/
	unsigned char	select_timeout;		/* timeout (R5)				*/
    } ifcfg;

    /* queue handling */
    struct {
    	Queue_t		issue;			/* issue queue				*/
    	Queue_t		disconnected;		/* disconnected command queue		*/
    } queues;

    /* per-device info */
    struct {
	unsigned char	disconnect_ok:1;	/* device can disconnect		*/
    } device[8];
    unsigned char	busyluns[8];		/* array of bits indicating LUNs busy	*/

    /* dma */
    struct {
	int	(*setup) (struct Scsi_Host *host, Scsi_Pointer *SCp, dmadir_t direction);
	int	(*pseudo)(struct Scsi_Host *host, Scsi_Pointer *SCp, dmadir_t direction, int transfer);
	void	(*stop)  (struct Scsi_Host *host, Scsi_Pointer *SCp);
    } dma;

    /* miscellaneous */
    int			internal_done;		/* flag to indicate request done */
} FAS216_Info;

/*
 * Function: int fas216_init (struct Scsi_Host *instance)
 *
 * Purpose : initialise FAS/NCR/AMD SCSI ic.
 *
 * Params  : instance - a driver-specific filled-out structure
 *
 * Returns : 0 on success
 */
extern int fas216_init (struct Scsi_Host *instance);

/*
 * Function: int fas216_abort (Scsi_Cmnd *SCpnt)
 *
 * Purpose : abort a command if something horrible happens.
 *
 * Params  : SCpnt - Command that is believed to be causing a problem.
 *
 * Returns : one of SCSI_ABORT_ macros.
 */
extern int fas216_abort (Scsi_Cmnd *);

/*
 * Function: int fas216_reset (Scsi_Cmnd *SCpnt, unsigned int reset_flags)
 *
 * Purpose : resets the adapter if something horrible happens.
 *
 * Params  : SCpnt - Command that is believed to be causing a problem.
 *	     reset_flags - flags indicating reset type that is believed to be required.
 *
 * Returns : one of SCSI_RESET_ macros, or'd with the SCSI_RESET_*_RESET macros.
 */
extern int fas216_reset (Scsi_Cmnd *, unsigned int);

/*
 * Function: int fas216_queue_command (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
 *
 * Purpose : queue a command for adapter to process.
 *
 * Params  : SCpnt - Command to queue
 *	     done  - done function to call once command is complete
 *
 * Returns : 0 - success, else error
 */
extern int fas216_queue_command (Scsi_Cmnd *, void (*done)(Scsi_Cmnd *));

/*
 * Function: int fas216_command (Scsi_Cmnd *SCpnt)
 *
 * Purpose : queue a command for adapter to process.
 *
 * Params  : SCpnt - Command to queue
 *
 * Returns : scsi result code
 */
extern int fas216_command (Scsi_Cmnd *);

/*
 * Function: void fas216_intr (struct Scsi_Host *instance)
 *
 * Purpose : handle interrupts from the interface to progress a command
 *
 * Params  : instance - interface to service
 */
extern void fas216_intr (struct Scsi_Host *instance);

/*
 * Function: int fas216_release (struct Scsi_Host *instance)
 *
 * Purpose : release all resources and put everything to bed for FAS/NCR/AMD SCSI ic.
 *
 * Params  : instance - a driver-specific filled-out structure
 *
 * Returns : 0 on success
 */
extern int fas216_release (struct Scsi_Host *instance);

#endif /* FAS216_H */
